/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for Petit FatFs (C)ChaN, 2014      */
/*-----------------------------------------------------------------------*/

// [2014/06/23] Modified by Eleken
#include "../main.h"

#include "diskio.h"

typedef struct _sCardInfo{
	uint8_t addressShift;
}sCardInfo;

static sCardInfo g_sCardInfo;

static void SPI_Write(uint8_t data){
	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}
static uint8_t SPI_Read(void){
	SPDR = 0xFF;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}

static void MMC_SendCmd(uint8_t cmd, uint32_t arg, uint8_t crc){
	
	SPDR = cmd;
	while(!(SPSR & (1<<SPIF)));
	SPDR = (uint8_t)(arg>>24);
	while(!(SPSR & (1<<SPIF)));
	SPDR = (uint8_t)(arg>>16);
	while(!(SPSR & (1<<SPIF)));
	SPDR = (uint8_t)(arg>>8);
	while(!(SPSR & (1<<SPIF)));
	SPDR = (uint8_t)(arg);
	while(!(SPSR & (1<<SPIF)));
	SPDR = crc;
	while(!(SPSR & (1<<SPIF)));
	
}
static uint8_t MMC_CheckResponse(void){
	uint16_t a=5000;
	while(a--){
		SPDR = 0xFF;
		while(!(SPSR & (1<<SPIF)));
		
		uint8_t d = SPDR;
		if (!(d & 0x80)){
			return d;
		}
	}
	return 0xFF;
}

/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (void)
{
	
	uint16_t i;
	uint8_t res;
	// dummy read
	SBI(MMC_PORT, MMC_CS);
	for(i=0;i<12;i++){
		SPI_Write(0xFF);
	}
	CBI(MMC_PORT, MMC_CS);
	
//	_delay_ms(1);
	MMC_SendCmd(0b01000000, 0, 0b10010101); // CMD0
	res = MMC_CheckResponse();
	if(res != 0x01){
		SBI(MMC_PORT, MMC_CS);
		if(res == 0xFF) return STA_NODISK; // 
		return STA_NOINIT;
	}
	
	uint8_t r8_accept = 0;
	
	
	MMC_SendCmd(0b01001000, 0x01AA, 0x87); // CMD8
	res = MMC_CheckResponse();

	uint32_t sdv2_res = 0;
	// Check R7 Response
	{
		uint32_t  d;
		for(i=0; i<4; i++){
			d = SPI_Read();
			sdv2_res = (sdv2_res << 8) | d;
		}
		/*
		sdv2_res |= (uint32_t) d << 24;
		d = SPI_Read();
		sdv2_res |= (uint32_t) d << 16;
		d = SPI_Read();
		sdv2_res |= (uint32_t) d << 8;
		d = SPI_Read();
		sdv2_res |= (uint32_t) d ;
		*/
	}
	
	
	if((sdv2_res & 0x00000FFF) == 0x01AA){ // SDv2
		r8_accept = 1;
	}
	
	while(1){
		MMC_SendCmd(0b01110111, 0, 1); // ACMD41
		res = MMC_CheckResponse();
		MMC_SendCmd(0b01101001, (r8_accept? 0x40000000: 0x00000000), 1); // ACMD41
		res = MMC_CheckResponse();
		if(res == 0x00) break; // correct
		else if(res != 0x01){
			SBI(MMC_PORT, MMC_CS);
			return STA_NOINIT;
		}
	}
	
	MMC_SendCmd(0b01010000, 512, 1); // CMD16
	res = MMC_CheckResponse();
	
	// CMD58
	if(r8_accept){ // SDv2
		while(1){
			MMC_SendCmd(0b01111010, 0, 1); // CMD58
			res = MMC_CheckResponse();
			if(res == 0){
				// Check R3 Response (OCR)
				uint8_t d;
				d = SPI_Read();
				if(d & 0x40) g_sCardInfo.addressShift = 0; // Block address
				else g_sCardInfo.addressShift = 9; // Byte address
				d = SPI_Read();
				d = SPI_Read();
				d = SPI_Read();
				break;
			}
			else if(res != 0x01){
				SBI(MMC_PORT, MMC_CS);
				return STA_NOINIT;
			}
		}
	}else {
		g_sCardInfo.addressShift = 9;
	}
	
	SBI(MMC_PORT, MMC_CS);
	SPI_Write(0xFF); // dummy write
	
	/*
	res = Read_FAT(MMC_buf); // read FAT informaiton
	if(res == 2) return STA_NOINIT; // unsupported format
	else if(res != 0) return STA_NOINIT;
	*/
	

	return 0;
}



/*-----------------------------------------------------------------------*/
/* Read Partial Sector                                                   */
/*-----------------------------------------------------------------------*/

DRESULT disk_readp (
	BYTE* buff,		/* Pointer to the destination object */
	DWORD sector,	/* Sector number (LBA) */
	UINT offset,	/* Offset in the sector */
	UINT count		/* Byte count (bit15:destination) */
)
{
	DRESULT result = RES_ERROR;
	
	uint16_t  i;
	CBI(PORTB, MMC_CS);
	
	if(g_sCardInfo.addressShift) sector = sector << 9;
	
	do{
		MMC_SendCmd(0b01010001, sector, 1); // CMD17
		
		// Read R1 Response
		uint8_t response = MMC_CheckResponse();
		if(response != 0) break;
		
		// Read Data Token
		for(i=0; i<20000; i++){ // Polling with roop limitation
			if(SPI_Read() == 0xFE) break;// data token start
		}
		if(i == 20000) {
			break; // (time-out)
		}
		
		// Read Data
		for(i=0; i<512; i++){
			SPI_Write(0xFF);
			if(i>=offset && i<offset+count){
				buff[i-offset] = SPDR;
			}
		}
		result = RES_OK;
		
	}while(0);
	
	SPI_Write(0xFF); // CRC byte
	SPI_Write(0xFF); // CRC byte

	SBI(PORTB, MMC_CS);
	SPI_Write(0xFF); // dummy write
	
	return result;
}



/*-----------------------------------------------------------------------*/
/* Write Partial Sector                                                  */
/*-----------------------------------------------------------------------*/

DRESULT disk_writep (
	const BYTE* buff,		/* Pointer to the data to be written, NULL:Initiate/Finalize write operation */
	DWORD sc		/* Sector number (LBA) or Number of bytes to send */
)
{
	DRESULT result = RES_ERROR;
	static uint16_t writeCnt;

	if (!buff) {
		if (sc) {
			do{
				CBI(PORTB, MMC_CS);
				// Initiate write process
				if(g_sCardInfo.addressShift) sc = sc << 9;
				MMC_SendCmd(0b01011000, sc, 1); // CMD24
				
				// Read R1 Response
				uint8_t response = MMC_CheckResponse();
				if(response != 0) break;
				
				SPI_Write(0xFF); // dummy write
				SPI_Write(0xFE); // data token start
				writeCnt = 512;
				result = RES_OK;
			}while(0);
		} else {
			uint16_t  i;
			// Finalize write process
			while(writeCnt--) SPI_Write(' '); // remain bytes
			SPI_Write(0xFF); // CRC byte
			SPI_Write(0xFF); // CRC byte
			
			// wait until write complete
			if((SPI_Read()&0x1F) == 0x05) { // write command was accepted
				for(i=0; i<20000; i++){ // Polling with roop limitation
					if(SPI_Read() == 0xFF) break; // Return from busy state
					_delay_us(100);
				}
				if(i != 20000) result = RES_OK;
			}
			
			SBI(PORTB, MMC_CS);
			SPI_Write(0xFF); // dummy write
		}
	} else {
		// Send data to the disk
		uint8_t sc8 = (uint8_t)sc;
		for(uint8_t i=0; i<sc8; i++){
			SPI_Write(buff[i]); // CRC byte
			writeCnt--;
		}
		result = RES_OK;
	}

	return result;
}

