// MSM62X42B (OKI) Driver Rutine
// 2014.06.23 Eleken
// rev_c1815@yahoo.co.jp
// See also: http://eleken.futene.net/

#include "main.h"
#include "msm62x42b.h"




// addr, data: 4-bit
static void MSM62X42B_Write(uint8_t addr, uint8_t data){
	SBI(PORTD, RTC_ALE);
	PORTD &= 0b00001111;
	PORTD |= (addr<<4);
	_delay_us(10);
	CBI(PORTD, RTC_ALE);
	_delay_us(10);
	
	PORTD &= 0b00001111;
	PORTD |= (data << 4);
	_delay_us(10);
	CBI(PORTB, RTC_WR);
	_delay_us(10);
	SBI(PORTB, RTC_WR);
}

static void MSM62X42B_Write8(uint8_t addr, uint8_t data){
	MSM62X42B_Write(addr+1, data>>4);
	MSM62X42B_Write(addr, data&0x0F);
}
// addr, data: 4-bit
static void MSM62X42B_Read(uint8_t addr, uint8_t* data){
	SBI(PORTD, RTC_ALE);
	PORTD &= 0b00001111;
	PORTD |= (addr<<4);
	_delay_us(10);
	CBI(PORTD, RTC_ALE);
	_delay_us(10);

	PORTD &= 0b00001111;
	DDRD  &= 0b00001111; // set D0-D3 to input
	CBI(PORTB, RTC_RD);
	_delay_us(9);
	*data = (PIND >> 4);
	SBI(PORTB, RTC_RD);
	_delay_us(1);

	DDRD  |= 0b11111100; // set D0-D3 to output
}
static void MSM62X42B_Read8(uint8_t addr, uint8_t* data){
	uint8_t temp;
	MSM62X42B_Read(addr+1, &temp);
	MSM62X42B_Read(addr, data);
	*data |= temp<<4;
}

// proc. about 0.2ms
void MSM62X42B_SetCurTime(const sTime* pCurTime){
	for(;;){
		uint8_t temp;
		MSM62X42B_Write(0x0D, 0x04 | 1); // HOLD=1
		MSM62X42B_Read(0x0D, &temp);
		if((temp & 0x02) == 0) break; // BUSY=0?
		MSM62X42B_Write(0x0D, 0x04 | 0); // HOLD=0
		_delay_us(62);
	}

	MSM62X42B_Write8(0x0A, pCurTime->year);
	MSM62X42B_Write8(0x08, pCurTime->month);
	MSM62X42B_Write8(0x06, pCurTime->day);
	MSM62X42B_Write8(0x04, pCurTime->hour);
	MSM62X42B_Write8(0x02, pCurTime->min);
	MSM62X42B_Write8(0x00, 0);
	
	MSM62X42B_Write(0x0D, 0x04 | 0); // HOLD=0
}

// proc. about 0.2ms
void MSM62X42B_GetCurTime(sTime* pCurTime){
	for(;;){
		uint8_t temp;
		MSM62X42B_Write(0x0D, 0x04 | 1); // HOLD=1
		MSM62X42B_Read(0x0D, &temp);
		if((temp & 0x02) == 0) break; // BUSY=0?
		MSM62X42B_Write(0x0D, 0x04 | 0); // HOLD=0
		_delay_us(62);
	}
	
	MSM62X42B_Read8(0x0A, &pCurTime->year);
	MSM62X42B_Read8(0x08, &pCurTime->month);
	MSM62X42B_Read8(0x06, &pCurTime->day);
	MSM62X42B_Read8(0x04, &pCurTime->hour);
	MSM62X42B_Read8(0x02, &pCurTime->min);
	MSM62X42B_Read8(0x00, &pCurTime->sec);
	
	MSM62X42B_Write(0x0D, 0x04 | 0); // HOLD=0
}

#ifdef RTC_INITPORT_DEF
// Call this function before Reset()
void MSM62X42B_InitPort(void){
	SBI(PORTC, RTC_STDP); // STD.P
	CBI(DDRC,  RTC_STDP);
	
	SBI(PORTB, RTC_RD); // RD
	SBI(PORTB, RTC_WR); // WR
	DDRB |= 0b00000011; // RD,WR
	
	DDRD |= 0b11111100; // D3-D0, CS1, ALE
	
	
	uint8_t i;
	for(i=0;i<100;i++) _delay_ms(10);
	SBI(PORTD, RTC_CS1);
	
}
#endif

// Power-on initialization process
// Note: since this process stops the internal clock temporary, power-on process of the MCU side should not call this function
//  in case of battery-backup system
void MSM62X42B_Reset(void){
	
	MSM62X42B_Write(0x0E, 0b00000100); // STD.P is output, the interval is 1 sec. ITRPT=0, MASK=0
	MSM62X42B_Write(0x0F, 0b00000111); // test=0, 24/12=1, reset=1, stop=1 
	
	MSM62X42B_Write(0x0F, 0b00000110); // test=0, 24/12=1, reset=0, stop=1 
	
	MSM62X42B_Write(0x0F, 0b00000100); // test=0, 24/12=1, reset=0, stop=0 
	MSM62X42B_Write(0x0D, 0x04 | 0); // HOLD=0
}


// Enable STD.P output / 1.0 sec interval pulse
void MSM62X42B_EnableOutput(void){
	MSM62X42B_Write(0x0E, 0b00000100); // STD.P is output, the interval is 1 sec. ITRPT=0, MASK=0
}
void MSM62X42B_DisableOutput(void){
	MSM62X42B_Write(0x0E, 0b00000011); // ITRPT=1, MASK=1
}

